import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

import torch
import torch.nn as nn
from torch.utils.data import DataLoader

from tqdm import tqdm

from TabularTransformer import TabularTransformer
from ctd_gan_networks import Critic, Generator
from ctd_clusterer import ctdClusterer


class ctdGAN:
    """
    ctdGAN implementation

    ctdGAN conditionally generates tabular data with the aim of confronting class imbalance. The model uses both
    cluster and class labels for training. It applies a cluster-aware data transformation mechanism and introduces
    a loss function that penalizes the generation of samples with incorrect cluster and class labels. New data
    instances are generated via a probabilistic sampling strategy.
    """

    def __init__(self, discriminator=(128, 128), generator=(256, 256), embedding_dim=128, epochs=300, batch_size=32,
                 pac=1, lr=2e-4, decay=1e-6, sampling_strategy='auto',
                 scaler='stds', cluster_method='kmeans', max_clusters=20, random_state=0):
        """
        ctdGAN initializer

        Args:
            discriminator (tuple): a tuple with number of neurons for each fully connected layer of the model's Critic.
                The tuple elements determine the dimensionality of the output of each layer.
            generator (tuple): a tuple with number of neurons for each fully connected layer of the model's Generator.
                The tuple elements determine the dimensionality of the output of each residual block of the Generator.
            embedding_dim (int): Size of the normally distributed latent vector passed to the Generator.
            epochs (int): The number of training epochs.
            batch_size (int): The number of data instances per training batch. Must be α multiple of `pac`.
            scaler (string): A descriptor that defines a transformation on the cluster's data. Values:

              * '`None`'  : No transformation takes place; the data is considered immutable
              * '`stds`'  : Standard scaler
              * '`mms01`' : Min-Max scaler in the range (0,1)
              * '`mms11`' : Min-Max scaler in the range (-1,1) - so that data is suitable for tanh activations
              * '`yeo`':  Yeo-Johnson Power Transformer
            pac (int): The number of samples to group together as input to the Critic.
            lr (real): The value of the learning rate parameter for the Generator/Critic Adam optimizers.
            decay (real): The value of the weight decay parameter for the Generator/Critic Adam optimizers.
            sampling_strategy (string or dictionary): How the model generates samples:

                - 'auto': balance the dataset by oversampling the minority classes.
                - 'balance-clusters': balance the dataset by balancing its clusters.
                - 'create-new': create a new dataset with the same class distribution as the one that was trained with.
                - dict: a dictionary that indicates the number of samples to be generated from each class
            max_clusters (int): The maximum number of clusters to create.
            random_state (int): Seed the random number generators. Use the same value for reproducible results.
        """
        self._name = "ctdGAN"

        self._n_classes = 0                     # Number of classes in the input dataset
        self._n_clusters = 0
        self._max_clusters = max_clusters

        self._input_dim = 0                     # Input data dimensionality
        self._embedding_dim = embedding_dim
        self._categorical_columns = []

        self._gen_samples_ratio = None          # Array [number of samples to generate per class]
        self._samples_per_class = None          # Array [ [x_train_per_class] ]

        self._device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

        self._pac = pac
        self._disc_lr = lr
        self._gen_lr = lr
        self._disc_decay = decay
        self._gen_decay = decay
        self._epochs = epochs                   # Number of training epochs
        self._batch_size = batch_size           # Number of data instances per training batch

        self._sampling_strategy = sampling_strategy  # Used in `fit_resample`: How GAN generates data

        # Discriminator parameters (object, architecture, optimizer)
        self.D_ = None
        self.D_Arch_ = discriminator
        self.D_optimizer_ = None

        # Generator parameters (object, architecture, optimizer)
        self.G_ = None
        self.G_Arch_ = generator
        self.G_optimizer_ = None

        self._cluster_method = cluster_method
        if scaler not in ('None', 'none', 'stds', 'mms01', 'mms11', 'yeo'):
            self._scaler = 'mms11'
        else:
            self._scaler = scaler

        # clustered_transformer performs clustering and data transformation.
        self._clustered_transformer = None

        # discrete_transformer performs dataset-wise one-hot-encoding of the categorical columns
        self._discrete_transformer = None
        self._random_state = random_state       # An integer to seed the random number generators

    @staticmethod
    def plot_losses(losses, store_losses):
        """
        Plot the Discriminator loss and the Generator loss values vs. the training epoch.
        Args:
            losses: A list of tuples (iteration, epoch, Discriminator loss, Generator loss) recorded during training.
            store_losses: The file path to store the plot.
        """
        custom_params = {'axes.facecolor': '#f0f0f0', 'axes.edgecolor': 'black', 'grid.color': '#ffffff',
                         'legend.labelspacing': 0.5, 'legend.fontsize': 15.0,
                         'axes.spines.right': False, 'axes.spines.top': False}
        sns.set_theme(rc=custom_params, font_scale=1.6)
        colors = ['#0072b0', '#ffa868']

        columns = ['Iteration', 'Epoch', 'Critic Loss it', 'Generator Loss it']
        df = pd.DataFrame(losses, columns=columns)
        df.to_csv(store_losses + 'ctdGAN_losses.csv', sep=';', decimal='.', index=False)

        df_mean = pd.DataFrame()
        df_mean['Critic Loss'] = df.groupby('Epoch')['Critic Loss it'].mean()
        df_mean['Generator Loss'] = df.groupby('Epoch')['Generator Loss it'].mean()
        df_mean['Epoch'] = df_mean.index + 1
        df_mean.to_csv(store_losses + 'ctdGAN_mean_losses.csv', sep=';', decimal='.', index=False)

        # plot = df.plot(x="Iteration", y=["Discriminator Loss", "Generator Loss"], ylim=(0, 1))
        plot = df_mean.plot(x='Epoch', y=['Critic Loss', 'Generator Loss'], kind='line', color=colors,
                            xlabel='Epoch', ylabel='Loss', title="Dry Bean")

        fig = plot.get_figure()
        # fig.savefig(store_losses + 'GAN_losses.png')
        fig.savefig(store_losses + 'GAN_losses.pdf', format='pdf', bbox_inches='tight')

        plt.show()

    @staticmethod
    def _gumbel_softmax(logits, tau=1.0, hard=False, eps=1e-10, dim=-1):
        """Deals with the instability of the gumbel_softmax for older versions of torch.

        For more details about the issue:
        https://drive.google.com/file/d/1AA5wPfZ1kquaRtVruCd6BiYZGcDeNxyP/view?usp=sharing

        Args:
            logits (array(…, num_features)): Un-normalized log probabilities
            tau: Non-negative scalar temperature/
            hard (bool): If True, the returned samples will be transformed to one-hot vectors,
                but will be differentiated as if it is the soft sample in autograd.
            dim (int): A dimension along which softmax will be computed. Default: -1.

        Returns:
            Sampled tensor of same shape as logits from the Gumbel-Softmax distribution.
        """
        for _ in range(10):
            transformed = nn.functional.gumbel_softmax(logits, tau=tau, hard=hard, eps=eps, dim=dim)
            if not torch.isnan(transformed).any():
                return transformed

        raise ValueError('gumbel_softmax returning NaN.')

    def _apply_activate(self, data):
        """Apply proper activation function to the output of the generator."""
        data_t = []
        st = 0
        for column_info in self._discrete_transformer.output_info_list:
            for span_info in column_info:
                if span_info.activation_fn == 'tanh':
                    ed = st + span_info.dim
                    data_t.append(torch.tanh(data[:, st:ed]))
                    st = ed
                elif span_info.activation_fn == 'softmax':
                    ed = st + span_info.dim
                    # transformed = torch.softmax(data[:, st:ed], dim=1)
                    transformed = self._gumbel_softmax(data[:, st:ed], tau=0.2)
                    data_t.append(transformed)
                    st = ed
                else:
                    raise ValueError(f'Unexpected activation function {span_info.activation_fn}.')

        return torch.cat(data_t, dim=1)

    def cluster_transform(self, x_train, y_train, categorical_columns):
        """
        Perform clustering and (optionally) transform the data in the generated clusters.

        Args:
            x_train: The training data instances.
            y_train: The classes of the training data instances.
            categorical_columns: The discrete columns in the dataset. The last 2 columns indicate the cluster and class.

        Returns:
            A tensor with the preprocessed data.
        """
        self._categorical_columns = list(categorical_columns)
        self._n_classes = len(set(y_train))
        self._input_dim = x_train.shape[1]
        continuous_columns = [c for c in range(self._input_dim) if c not in self._categorical_columns]

        # ====== Initialize and fit the Clustered Transformer object that: i) partitions the real space, and
        # ====== ii) performs data transformations (scaling, PCA, outlier detection, etc.)
        self._samples_per_class = np.unique(y_train, return_counts=True)[1]

        self._clustered_transformer = ctdClusterer(cluster_method=self._cluster_method,
                                                   max_clusters=self._max_clusters,
                                                   scaler=self._scaler,
                                                   samples_per_class=self._samples_per_class,
                                                   continuous_columns=tuple(continuous_columns),
                                                   discrete_columns=tuple(self._categorical_columns),
                                                   embedding_dim=self._embedding_dim,
                                                   random_state=self._random_state)

        train_data = self._clustered_transformer.perform_clustering(x_train, y_train, self._n_classes, self._pac)

        self._n_clusters = self._clustered_transformer.num_clusters_

        # ====== Append the cluster and class labels to the collection of discrete columns
        self._categorical_columns.append(self._input_dim)
        self._categorical_columns.append(self._input_dim + 1)

        # ====== Transform the discrete columns only; the continuous columns have been scaled at cluster-level.
        self._discrete_transformer = TabularTransformer(cont_normalizer='None', clip=False)
        self._discrete_transformer.fit(train_data, self._categorical_columns)

        ret_data = self._discrete_transformer.transform(train_data)

        return ret_data

    def sample_latent_space(self, num_samples):
        """Latent space sampler

        Samples the latent space and returns the latent feature vectors z, the one-hot-encoded cluster labels and
        the one-hot-encoded class labels.

        Args:
            num_samples: The number of latent data instances.

        Returns:
             latent continuous variables, latent distinct variables
        """

        # === Discrete variables - Simple random one-hot-encoded integers
        col = 0
        latent_disc = []
        for column_metadata in self._discrete_transformer.output_info_list:
            for span_info in column_metadata:
                col += 1
                if span_info.activation_fn == 'softmax':
                    col_length = span_info.dim
                    random_numbers = torch.randint(low=0, high=col_length, size=(1, num_samples))[0]
                    z_disc = nn.functional.one_hot(random_numbers, num_classes=col_length)
                    latent_disc.append(z_disc)

        latent_disc = torch.hstack(latent_disc).to(self._device)

        # === Continuous variables - Sample from Normal distribution.
        mean = torch.zeros(num_samples, self._embedding_dim)
        std = mean + 1
        latent_cont = torch.normal(mean=mean, std=std).to(self._device)

        return latent_cont, latent_disc

    def generator_loss(self, critic_output, latent_disc_data, generated_data):
        """Generator loss function

        The loss function of the Generator combines the Discriminator Loss, and the mis-predictions of
        the cluster and class labels.

        Args:
            critic_output: The output of the Discriminator.
            latent_disc_data: The discrete latent data where the Generator sampled from.
            generated_data: The data generated by the Generator.
        """
        num_generated_samples = generated_data.size()[0]

        cluster_st_idx = self._discrete_transformer.output_dimensions - self._n_clusters - self._n_classes
        cluster_ed_idx = cluster_st_idx + self._n_clusters
        # print("Cluster st:", cluster_st_idx, ", Cluster ed:", cluster_ed_idx)

        class_st_idx = self._discrete_transformer.output_dimensions - self._n_classes
        class_ed_idx = class_st_idx + self._n_classes
        # print("Class st:", class_st_idx, ", Class ed:", class_ed_idx)

        # Loss for the discrete columns - Including the class & cluster labels.
        # print("Real data shape", real_data.shape, " - Latent data shape", latent_data.shape)
        discrete_loss = []
        st_idx = 0
        st_disc_idx = 0
        for column_metadata in self._discrete_transformer.output_info_list:
            for span_info in column_metadata:
                if len(column_metadata) != 1 or span_info.activation_fn != 'softmax':
                    # Continuous column
                    st_idx += span_info.dim
                else:
                    ed_idx = st_idx + span_info.dim
                    ed_disc_idx = st_disc_idx + span_info.dim

                    # print("\t== Latent Discrete Data:\nStart:", st_disc_idx, ", End:", ed_disc_idx, "\nData:\n",
                    #       latent_disc_data[:, st_disc_idx:ed_disc_idx])
                    # print("\t== Generated Discrete Data:\nStart:", st_idx, ", End:", ed_idx, "\nData:\n",
                    #       generated_data[:, st_idx:ed_idx])

                    gen_d = generated_data[:, st_idx:ed_idx]
                    gen_c = torch.argmax(gen_d, dim=1)

                    lat_d = latent_disc_data[:, st_disc_idx:ed_disc_idx].to(dtype=torch.float)
                    lat_c = torch.argmax(lat_d, dim=1)

                    # Penalize mis-clusters more heavily and in a dynamic manner
                    if st_idx == cluster_st_idx and ed_idx == cluster_ed_idx:
                        # print("Lat Clusters:\n", lat_d, "(", lat_c, ")", "\nGen Clusters:\n", gen_d, "(", gen_c, ")")

                        mis_clustered = np.sum([1 for i in range(num_generated_samples) if gen_c[i] != lat_c[i]])
                        beta = 1.0 + mis_clustered / num_generated_samples
                        # beta = 1.0

                        if self._n_clusters == 2:
                            tmp = beta * nn.functional.binary_cross_entropy(gen_d, lat_d, reduction='none')[:, 0]
                        else:
                            tmp = beta * nn.functional.cross_entropy(gen_d, lat_c, reduction='none')
                        # tmp = torch.zeros((num_generated_samples, )).to(self._device)
                        # print(beta)

                    # Penalize mis-classifications more heavily
                    elif st_idx == class_st_idx and ed_idx == class_ed_idx:
                        # self._categorical_columns = 1 + number of discrete columns (2 is from the class + cluster col)
                        # gamma = len(self._categorical_columns) - 1
                        gamma = 1.0

                        if self._n_classes == 2:
                            tmp = gamma * nn.functional.binary_cross_entropy(gen_d, lat_d, reduction='none')[:, 0]
                        else:
                            tmp = gamma * nn.functional.cross_entropy(gen_d, lat_c, reduction='none')
                    else:
                        tmp = nn.functional.cross_entropy(gen_d, lat_c, reduction='none')

                    discrete_loss.append(tmp)

                    st_idx = ed_idx
                    st_disc_idx = ed_disc_idx

        cond_loss = torch.stack(discrete_loss, dim=1)
        discrete_columns_loss = cond_loss.sum() / num_generated_samples
        # print("discrete_columns_loss = ", discrete_columns_loss)

        gen_loss = -torch.mean(critic_output) + discrete_columns_loss

        return gen_loss

    def discriminator_loss(self, real_data, generated_data):
        """Discriminator (Critic) loss function

        The loss function of the Critic measures the difference in the quality of the real and generated data.

        Args:
            real_data: The real data in the batch.
            generated_data: The data generated by the Generator.
        """

        real_data_quality = torch.mean(self.D_(real_data))
        generated_data_quality = torch.mean(self.D_(generated_data))

        return -(real_data_quality - generated_data_quality)

    def train_batch(self, real_data):
        """
        Given a batch of input data, `train_batch` updates the Critic and Generator weights using the respective
        optimizers and back propagation.

        Args:
            real_data: data for ctdGAN training: a batch of concatenated sample vectors + one-hot-encoded class vectors.
        """

        # If the size of the batch does not allow an organization of the input vectors in packs of size self._pac, then
        # abort silently and return without updating the model parameters.
        num_samples = real_data.shape[0]
        if num_samples % self._pac != 0:
            print("pac error")
            return 0, 0

        real_data = real_data.to(torch.float).to(self._device)

        # DISCRIMINATOR TRAINING
        with torch.no_grad():
            # Take samples from the latent space
            latent_cont, latent_disc = self.sample_latent_space(num_samples)
            latent_data = torch.cat((latent_cont, latent_disc), dim=1)

        # Pass the latent samples through the Generator and generate fake samples.
        # Apply a dual activation function: tanh for the continuous columns, softmax for the discrete ones.
        generated_data = self._apply_activate(self.G_(latent_data))

        # Pass the mixed data to the Discriminator and train the Discriminator (update its weights with backprop).
        # The loss function quantifies the Discriminator's ability to classify a real/fake sample as real/fake.
        pen = self.D_.calc_gradient_penalty(real_data, generated_data, device=self._device)
        disc_loss = self.discriminator_loss(real_data, generated_data)

        self.D_optimizer_.zero_grad(set_to_none=True)
        pen.backward(retain_graph=True)
        disc_loss.backward()
        self.D_optimizer_.step()

        # GENERATOR TRAINING
        # Sample data from the latent spaces
        with torch.no_grad():
            latent_cont, latent_disc = self.sample_latent_space(num_samples)
            latent_data = torch.cat((latent_cont, latent_disc), dim=1)

        # Pass the latent data through the Generator to synthesize samples
        generated_data = self._apply_activate(self.G_(latent_data))

        # Compute and back propagate the Generator loss
        d_predictions = self.D_(generated_data)
        gen_loss = self.generator_loss(d_predictions, latent_disc, generated_data)

        self.G_optimizer_.zero_grad(set_to_none=True)
        gen_loss.backward()
        self.G_optimizer_.step()

        return disc_loss, gen_loss

    def _train(self, x_train, y_train, categorical_columns=(), store_losses=None):
        """
        ctdGAN training process. The Generator and the Critic are trained jointly in the traditional adversarial
        fashion by optimizing `loss_function`.

        Args:
            x_train (NumPy array): The training data instances.
            y_train (NumPy array): The classes of the training data instances.
            categorical_columns: The columns to be considered as categorical
            store_losses: The file path where the values of the Discriminator and Generator loss functions are stored.
        """

        # Modify the size of the batch to align with self._pac
        factor = self._batch_size // self._pac
        batch_size = factor * self._pac

        # Prepare the data for training (Clustering, Computation of Probability Distributions, Transformations, etc.)
        training_data = self.cluster_transform(x_train, y_train, categorical_columns=categorical_columns)

        train_dataloader = DataLoader(training_data, batch_size=batch_size, shuffle=True)

        real_space_dimensions = self._discrete_transformer.output_dimensions
        latent_space_dimensions = self._embedding_dim + self._discrete_transformer.ohe_dimensions

        self.D_ = Critic(input_dim=real_space_dimensions, discriminator_dim=self.D_Arch_,
                         pac=self._pac).to(self._device)

        self.G_ = Generator(embedding_dim=latent_space_dimensions, architecture=self.G_Arch_,
                            data_dim=real_space_dimensions).to(self._device)

        self.D_optimizer_ = torch.optim.Adam(self.D_.parameters(),
                                             lr=self._disc_lr, weight_decay=self._disc_decay, betas=(0.5, 0.9))
        self.G_optimizer_ = torch.optim.Adam(self.G_.parameters(),
                                             lr=self._gen_lr, weight_decay=self._gen_decay, betas=(0.5, 0.9))

        losses = []
        it = 0
        for epoch in tqdm(range(self._epochs), desc="   Training..."):
            for real_data in train_dataloader:
                if real_data.shape[0] > 1:
                    disc_loss, gen_loss = self.train_batch(real_data)

                    if store_losses is not None:
                        it += 1
                        losses.append((it, epoch + 1, disc_loss.item(), gen_loss.item()))

        if store_losses is not None:
            self.plot_losses(losses, store_losses)

    def fit(self, x_train, y_train):
        """Invokes the GAN training process.

        Args:
            x_train: The training data instances.
            y_train: The classes of the training data instances.
        """
        self._train(x_train, y_train)

    def sample(self, num_samples, y=None, u=None):
        """ Create artificial samples using the GAN's Generator.

        Args:
            num_samples (int): The number of samples to generate.
            y (int): A condition on the class of the generated samples.
            u (int): A condition on the cluster of the generated samples.

        Returns:
            Artificial data instances created by the Generator.
        """
        num_columns = len(self._discrete_transformer.output_info_list)
        column_transform_info_list = self._discrete_transformer.get_column_transform_info_list()

        num_generated_samples, num_retries, max_retries = 0, 0, 100
        reconstructed_samples = []

        # Keep generating samples, until we reach the requested number of num_samples
        while num_generated_samples < num_samples:
            num_retries += 1

            # If no specific class is requested, select random class labels.
            if y is None:
                latent_classes = np.random.randint(low=0, high=self._n_classes, size=num_samples)
            # Otherwise, fill the classes tensor with the requested class (y) value
            else:
                latent_classes = np.full(shape=num_samples, fill_value=y)

            # We will determine the appropriate clusters later, according to the classes of the samples
            latent_clusters = np.zeros(shape=num_samples)

            # Select random values for the discrete variables. These values will be later one-hot-encoded.
            latent_disc = []
            col = 0
            column_labels = []
            for column_metadata in self._discrete_transformer.output_info_list:
                col = col + 1
                for span_info in column_metadata:
                    if span_info.activation_fn == 'softmax':
                        column_labels.append(str(col - 1))

                        # Discrete variables excluding the two last columns (i.e. the cluster and class labels)
                        if col < num_columns - 1:
                            col_length = span_info.dim
                            random_discrete_vals = np.random.randint(low=0, high=col_length, size=num_samples)
                            latent_disc.append(random_discrete_vals)

            # For each sample with a specific class, pick a random cluster with probability determined by the
            # corresponding p_matrix. In the same time, sample the probability distribution of each cluster to
            # get the latent representation of the continuous variables.
            latent_clusters_objs = []
            for s in range(num_samples):
                lat_class = int(latent_classes[s])
                p_matrix = self._clustered_transformer.probability_matrix_[lat_class]

                # Select the cluster with probability coming from the probability matrix of ctdClusterer
                if u is None:
                    latent_clusters[s] = np.random.choice(
                        a=np.arange(self._n_clusters, dtype=int), size=None, replace=True, p=p_matrix)
                else:
                    latent_clusters[s] = u

                latent_cluster_object = self._clustered_transformer.get_cluster(int(latent_clusters[s]))
                latent_clusters_objs.append(latent_cluster_object)

            # Put all discrete variables together into the same matrix (including the class and cluster labels)
            latent_disc.append(latent_clusters)
            latent_disc.append(latent_classes)
            latent_disc = pd.DataFrame(np.stack(latent_disc, axis=1), columns=column_labels)

            # Now one-hot-encode the discrete variables by using the OneHotEncoders that were used during training
            latent_disc_ohe = []
            for column_transform_info in column_transform_info_list:
                if column_transform_info.column_type != 'continuous':
                    column_name = column_transform_info.column_name
                    data = latent_disc[[column_name]]
                    one_hot_data = self._discrete_transformer.transform_discrete(column_transform_info, data)
                    latent_disc_ohe.append(one_hot_data)

            # Create the discrete and continuous tensors.
            latent_disc_ohe = torch.tensor(np.hstack(latent_disc_ohe))

            mean = torch.zeros(num_samples, self._embedding_dim)
            std = mean + 1
            latent_cont = torch.normal(mean=mean, std=std)

            # Concatenate the continuous with the discrete variables
            latent_data = torch.cat((latent_cont, latent_disc_ohe), dim=1).to(self._device)

            # Generate samples by passing the latent data to Generator
            generated_data = self._apply_activate(self.G_(latent_data)).cpu().detach().numpy()
            generated_samples = self._discrete_transformer.inverse_transform(generated_data)
            # print("\n\nLatent Classes:\n", latent_classes)
            # print("\n\nLatent Data:\n", latent_data)
            # print("\n\nGenerated Data:\n", generated_data)
            # print("\n\nGenerated Samples:\n", generated_samples)

            # Reverse the transformation of the generated samples. First inverse the transformation of the
            # continuous variables that have been encoded according to the cluster the sample belongs.
            for s in range(num_samples):
                z = generated_samples[s].reshape(1, -1)
                generated_class = z[0, z.shape[1]-1]

                if generated_class == latent_classes[s]:
                    num_generated_samples += 1
                    if num_generated_samples > num_samples:
                        return_samples = np.vstack(reconstructed_samples)
                        print("\t\t\tPerfectly created ", return_samples.shape, "samples from class", y)
                        return return_samples
                    reconstructed_sample = latent_clusters_objs[s].inverse_transform(z)
                    reconstructed_samples.append(reconstructed_sample)
                # print("Sample", s, "- Gen:", z, " ===>", reconstructed_sample)

            # If the maximum number of attempts has been exhausted, then exit the loop.
            # We will be generating fewer than the requested samples.
            if num_retries > max_retries:
                # If no sample has been retrieved, keep the last num samples
                if len(reconstructed_samples) == 0:
                    # print("\t\t\tI did not retrieve any results from class", y)
                    for s in range(num_samples):
                        z = generated_samples[s].reshape(1, -1)
                        reconstructed_sample = latent_clusters_objs[s].inverse_transform(z)
                        reconstructed_samples.append(reconstructed_sample)
                break

        return_samples = np.vstack(reconstructed_samples)
        print("\t\t\tIncompletely Created ", return_samples.shape, "samples from class", y)
        return return_samples

    def fit_resample(self, x_train, y_train, categorical_columns=()):
        """`fit_resample` alleviates the problem of class imbalance in imbalanced datasets. The function renders ctdGAN
        compatible with the `imblearn`'s interface, allowing its usage in over-sampling/under-sampling pipelines.

        In the `fit` part, the input dataset is used for training.
        In the `resample` part, the model generates synthetic data according to the value of `self._sampling_strategy`:

        - 'auto': balance the dataset by oversampling the minority classes.
        - 'balance-clusters': balance the dataset by balancing its clusters.
        - 'create-new': create a new dataset with the same class distribution as the one that was trained with
        - dict: a dictionary that indicates the number of samples to be generated from each class

        Args:
            x_train: The training data instances.
            y_train: The classes of the training data instances.
            categorical_columns: The columns to be considered as categorical

        Returns:
            x_resampled: The training data instances + the generated data instances.
            y_resampled: The classes of the training data instances + the classes of the generated data instances.
        """

        # Train ctdGAN with the input data
        # self.train(x_train, y_train, categorical_columns=categorical_columns, store_losses=paths.output_path_loss)
        self._train(x_train, y_train, categorical_columns=categorical_columns, store_losses=None)

        x_resampled = np.copy(x_train)
        y_resampled = np.copy(y_train)

        # auto mode: Use ctdGAN to equalize the number of samples per class. This is achieved by generating samples
        # of the minority classes (i.e. we perform oversampling).
        if self._sampling_strategy == 'auto':
            majority_class = np.array(self._samples_per_class).argmax()
            num_majority_samples = np.max(np.array(self._samples_per_class))

            # Perform oversampling
            for cls in tqdm(range(self._n_classes), desc="   Sampling..."):
                if cls != majority_class:
                    samples_to_generate = num_majority_samples - self._samples_per_class[cls]

                    if samples_to_generate > 1:
                        # Generate the appropriate number of samples to equalize cls with the majority class.
                        generated_samples = self.sample(num_samples=samples_to_generate, y=cls)
                        generated_classes = np.full(generated_samples.shape[0], cls)

                        x_resampled = np.vstack((x_resampled, generated_samples))
                        y_resampled = np.hstack((y_resampled, generated_classes))

        if self._sampling_strategy == 'balance-clusters':
            majority_class = np.array(self._samples_per_class).argmax()
            imb_matrix = self._clustered_transformer.imbalance_matrix_

            # Perform oversampling by performing cluster-based oversampling
            majority_samples = np.max(imb_matrix, axis=0)
            majority_classes = np.argmax(imb_matrix, axis=0)
            # print(imb_matrix)
            # print(majority_samples)
            # print(majority_classes)

            for u in tqdm(range(self._n_clusters), desc="   Sampling..."):
                # print("Cluster", u)
                for cls in range(self._n_classes):
                    ir = imb_matrix[cls][u] / majority_samples[u]

                    if cls != majority_classes[u] and cls != majority_class and ir > 0.01:
                        samples_to_generate = int(majority_samples[u] - imb_matrix[cls][u])
                        # print("\tI will create", samples_to_generate, "samples from Class", cls)

                        if samples_to_generate > 1:
                            # Generate the appropriate number of samples to equalize cls with the majority class.
                            generated_samples = self.sample(num_samples=samples_to_generate, y=cls, u=u)

                            if generated_samples is not None and generated_samples.shape[0] > 0:
                                # print("\t\tCreated", generated_samples.shape[0], "samples")
                                generated_classes = np.full(generated_samples.shape[0], cls)

                                x_resampled = np.vstack((x_resampled, generated_samples))
                                y_resampled = np.hstack((y_resampled, generated_classes))

        # dictionary mode: the keys correspond to the targeted classes. The values correspond to the desired number of
        # samples for each targeted class.
        elif isinstance(self._sampling_strategy, dict):
            for cls in tqdm(self._sampling_strategy, desc="   Sampling..."):
                # In imblearn sampling strategy stores the class distribution of the output dataset. So we have to
                # create the half number of samples, and we divide by 2.
                samples_to_generate = int(self._sampling_strategy[cls] / 2)

                # Generate the appropriate number of samples to equalize cls with the majority class.
                generated_samples = self.sample(num_samples=samples_to_generate, y=cls)

                if generated_samples is not None and generated_samples.shape[0] > 0:
                    # print("\t\tCreated", generated_samples.shape[0], "samples")
                    generated_classes = np.full(generated_samples.shape[0], cls)

                    x_resampled = np.vstack((x_resampled, generated_samples))
                    y_resampled = np.hstack((y_resampled, generated_classes))

        elif self._sampling_strategy == 'create-new':
            x_resampled = None
            y_resampled = None

            s = 0
            for cls in tqdm(range(self._n_classes), desc="   Sampling..."):
                # Generate as many samples, as the corresponding class cls
                samples_to_generate = int(self._samples_per_class[cls])
                generated_samples = self.sample(num_samples=samples_to_generate, y=cls)

                if generated_samples is not None and generated_samples.shape[0] > 0:
                    # print("\t\tCreated", generated_samples.shape[0], "samples")
                    generated_classes = np.full(generated_samples.shape[0], cls)

                    if s == 0:
                        x_resampled = generated_samples
                        y_resampled = generated_classes
                        s = 1
                    else:
                        x_resampled = np.vstack((x_resampled, generated_samples))
                        y_resampled = np.hstack((y_resampled, generated_classes))

        return x_resampled, y_resampled



# Tabular Transformer object: based on the Data Transformer of ctGAN
# Forked and modified from https://github.com/sdv-dev/CTGAN

import numpy as np
import pandas as pd

from sklearn.preprocessing import StandardScaler, MinMaxScaler, PowerTransformer
from sklearn.decomposition import PCA
from sklearn.pipeline import Pipeline

from joblib import Parallel, delayed
from rdt.transformers import ClusterBasedNormalizer, OneHotEncoder

from collections import namedtuple

SpanInfo = namedtuple('SpanInfo', ['dim', 'activation_fn'])
ColumnTransformInfo = namedtuple(
    'ColumnTransformInfo', [
        'column_name', 'column_type', 'transform', 'column_max', 'column_min', 'output_info', 'output_dimensions'
    ]
)

# TabularTransformer
class TabularTransformer(object):
    """Data Transformer.

    Model continuous columns with a BayesianGMM and normalized to a scalar [0, 1] and a vector.
    Discrete columns are encoded using a scikit-learn OneHotEncoder.
    """
    def __init__(self, cont_normalizer='None', max_clusters=10, weight_threshold=0.005, with_mean=True, with_std=True,
                 clip=False):
        """Create a data transformer.

        Args:
            cont_normalizer: Normalizer for the continuous columns:

              * '`None`'     : No transformation on the continuous columns takes place; the data is considered immutable
              * '`vgm`'      : Variational Gaussian Mixture + One-hot-encoded component labels.
              * '`stds`'     : Standard scaler
              * '`mms01`'    : Min-Max scaler in the range (0,1)
              * '`mms11`'    : Min-Max scaler in the range (-1,1) - so that data is suitable for tanh activations
              * '`stds-pca`' : Standard scaler & Principal Component Analysis
              * '`yeo`'      : Yeo-Johnson power transformer.

            max_clusters: Max number of Gaussian distributions in Bayesian GMM; used when `cont_normalizer='vgm'`
            weight_threshold: Weight threshold for a Gaussian distribution to be kept; used when `cont_normalizer='vgm'`
            with_mean: If True, it centers the data before scaling. This does not work (and will raise an exception)
                when attempted on sparse matrices, because centering them entails building a dense matrix which in
                common use cases is likely to be too large to fit in memory; used when `cont_normalizer='stds'`.
            with_std: If `True`, scale the data to unit variance (or equivalently, unit standard deviation);
                used when `cont_normalizer='stds'`.
            clip: If 'True' the reconstructed data will be clipped to their original minimum and maximum values.
        """
        self._cont_normalizer = cont_normalizer
        self._max_clusters = max_clusters
        self._weight_threshold = weight_threshold
        self._with_mean = with_mean
        self._with_std = with_std
        self._column_raw_dtypes = []
        self._column_transform_info_list = []
        self._clip = clip
        self.output_info_list = []
        self.output_dimensions = 0
        self.ohe_dimensions = 0
        self.dataframe = True

    def _fit_continuous(self, data):
        """Train Bayesian GMM for continuous columns.

        Args:
            data (pd.DataFrame): A dataframe containing a column.

        Returns:
            namedtuple: A ``ColumnTransformInfo`` object.
        """
        column_name = data.columns[0]
        max_val = np.array(data.max(axis=0))[0]
        min_val = np.array(data.min(axis=0))[0]

        cti = None
        if self._cont_normalizer == 'vgm':
            tran = ClusterBasedNormalizer(model_missing_values=True, max_clusters=min(len(data), self._max_clusters))
            tran.fit(data, column_name)
            num_components = sum(tran.valid_component_indicator)

            cti = ColumnTransformInfo(column_name=column_name, column_type='continuous', transform=tran,
                                      column_max=max_val, column_min=min_val,
                                      output_info=[SpanInfo(1, 'tanh'), SpanInfo(num_components, 'softmax')],
                                      output_dimensions=1 + num_components)

        elif self._cont_normalizer == 'stds':
            tran = StandardScaler(with_std=self._with_std, with_mean=self._with_mean)
            tran.fit(data)

            cti = ColumnTransformInfo(column_name=column_name, column_type='continuous', transform=tran,
                                      column_max=max_val, column_min=min_val,
                                      output_info=[SpanInfo(1, 'tanh')], output_dimensions=1)

        elif self._cont_normalizer == 'mms01':
            tran = MinMaxScaler(feature_range=(0, 1))
            tran.fit(data)

            cti = ColumnTransformInfo(column_name=column_name, column_type='continuous', transform=tran,
                                      column_max=max_val, column_min=min_val,
                                      output_info=[SpanInfo(1, 'tanh')], output_dimensions=1)

        elif self._cont_normalizer == 'mms11':
            tran = MinMaxScaler(feature_range=(-1, 1))
            tran.fit(data)

            cti = ColumnTransformInfo(column_name=column_name, column_type='continuous', transform=tran,
                                      column_max=max_val, column_min=min_val,
                                      output_info=[SpanInfo(1, 'tanh')], output_dimensions=1)

        elif self._cont_normalizer == 'stds-pca':
            tran = Pipeline(steps=[("std", StandardScaler(with_std=self._with_std, with_mean=self._with_mean)),
                                   ("pca", PCA())])
            tran.fit(data)

            cti = ColumnTransformInfo(column_name=column_name, column_type='continuous', transform=tran,
                                      column_max=max_val, column_min=min_val,
                                      output_info=[SpanInfo(1, 'tanh')], output_dimensions=1)

        elif self._cont_normalizer == 'yeo':
            tran = PowerTransformer(method='yeo-johnson', standardize=True)
            tran.fit(data)

            cti = ColumnTransformInfo(column_name=column_name, column_type='continuous', transform=tran,
                                      column_max=max_val, column_min=min_val,
                                      output_info=[SpanInfo(1, 'tanh')], output_dimensions=1)

        elif self._cont_normalizer == 'None':
            cti = ColumnTransformInfo(column_name=column_name, column_type='continuous', transform=None,
                                      column_max=max_val, column_min=min_val,
                                      output_info=[SpanInfo(1, 'tanh')], output_dimensions=1)

        return cti

    def _fit_discrete(self, data):
        """Fit one hot encoder for discrete column.

        Args:
            data (pd.DataFrame): A dataframe containing a column.

        Returns:
            namedtuple: A ``ColumnTransformInfo`` object.
        """
        column_name = data.columns[0]
        ohe = OneHotEncoder()
        ohe.fit(data, column_name)
        num_categories = len(ohe.dummies)
        self.ohe_dimensions += num_categories

        return ColumnTransformInfo(
            column_name=column_name, column_type='discrete', transform=ohe,
            column_max=-1, column_min=-1,
            output_info=[SpanInfo(num_categories, 'softmax')], output_dimensions=num_categories)

    def fit(self, raw_data, discrete_columns=()):
        """Fit the ``DataTransformer`` in a column-wise fashion. One transformer is fitted per column.

        Fits a ``ClusterBasedNormalizer`` for continuous columns and a ``OneHotEncoder`` for discrete columns. This
        step also counts the #columns in matrix data and span information.
        """
        self.output_info_list = []
        self.output_dimensions = 0
        self.dataframe = True

        if not isinstance(raw_data, pd.DataFrame):
            self.dataframe = False

            # work around for RDT issue #328 Fitting with numerical column names fails
            discrete_columns = [str(column) for column in discrete_columns]
            column_names = [str(num) for num in range(raw_data.shape[1])]
            raw_data = pd.DataFrame(raw_data, columns=column_names)

        self._column_raw_dtypes = raw_data.infer_objects().dtypes
        self._column_transform_info_list = []
        for column_name in raw_data.columns:
            if column_name in discrete_columns:
                column_transform_info = self._fit_discrete(raw_data[[column_name]])
            else:
                column_transform_info = self._fit_continuous(raw_data[[column_name]])

            self.output_info_list.append(column_transform_info.output_info)
            self.output_dimensions += column_transform_info.output_dimensions
            self._column_transform_info_list.append(column_transform_info)

    def _transform_continuous(self, column_transform_info, data):
        output = None
        if self._cont_normalizer == 'vgm':
            column_name = data.columns[0]
            flattened_column = data[column_name].to_numpy().flatten()
            data = data.assign(**{column_name: flattened_column})
            gm = column_transform_info.transform
            transformed = gm.transform(data)

            # Converts the transformed data to the appropriate output format. The first column (ending in '.normalized')
            # stays the same, but the label encoded column (ending in '.component') is one hot encoded.
            output = np.zeros((len(transformed), column_transform_info.output_dimensions))
            output[:, 0] = transformed[f'{column_name}.normalized'].to_numpy()
            index = transformed[f'{column_name}.component'].to_numpy().astype(int)
            output[np.arange(index.size), index + 1] = 1.0

        elif (self._cont_normalizer == 'stds' or self._cont_normalizer == 'mms01' or self._cont_normalizer == 'mms11'
              or self._cont_normalizer == 'stds-pca' or self._cont_normalizer == 'yeo'):
            output = column_transform_info.transform.transform(data)

        elif self._cont_normalizer == 'None':
            return data

        return output

    def transform_discrete(self, column_transform_info, data):
        ohe = column_transform_info.transform
        return ohe.transform(data).to_numpy()

    def _synchronous_transform(self, raw_data, column_transform_info_list):
        """Take a Pandas DataFrame and transform columns synchronous.

        Outputs a list with Numpy arrays.
        """
        column_data_list = []
        for column_transform_info in column_transform_info_list:
            column_name = column_transform_info.column_name
            data = raw_data[[column_name]]

            if column_transform_info.column_type == 'continuous':
                column_data_list.append(self._transform_continuous(column_transform_info, data))
            else:
                column_data_list.append(self.transform_discrete(column_transform_info, data))

        return column_data_list

    def _parallel_transform(self, raw_data, column_transform_info_list):
        """Take a Pandas DataFrame and transform columns in parallel.

        Outputs a list with Numpy arrays.
        """
        processes = []
        for column_transform_info in column_transform_info_list:
            column_name = column_transform_info.column_name
            data = raw_data[[column_name]]

            if column_transform_info.column_type == 'continuous':
                process = delayed(self._transform_continuous)(column_transform_info, data)
            else:
                process = delayed(self.transform_discrete)(column_transform_info, data)
            processes.append(process)

        return Parallel(n_jobs=-1)(processes)
        # return Parallel(n_jobs=0.5 * multiprocessing.cpu_count())(processes)

    def transform(self, raw_data):
        """Take raw data and output a matrix data."""
        if not isinstance(raw_data, pd.DataFrame):
            column_names = [str(num) for num in range(raw_data.shape[1])]
            raw_data = pd.DataFrame(raw_data, columns=column_names)

        # Only use parallelization with larger data sizes. Otherwise, the transformation will be slower.
        if raw_data.shape[0] < 500:
            column_data_list = self._synchronous_transform(raw_data, self._column_transform_info_list)
        else:
            column_data_list = self._parallel_transform(raw_data, self._column_transform_info_list)

        return np.concatenate(column_data_list, axis=1).astype(float)

    def _inverse_transform_continuous(self, column_transform_info, column_data, sigmas, st):
        ret_data = None
        encoder = column_transform_info.transform

        if self._cont_normalizer == 'vgm':
            data = pd.DataFrame(column_data[:, :2], columns=list(encoder.get_output_sdtypes()))
            data[data.columns[1]] = np.argmax(column_data[:, 1:], axis=1)
            if sigmas is not None:
                selected_normalized_value = np.random.normal(data.iloc[:, 0], sigmas[st])
                data.iloc[:, 0] = selected_normalized_value

            ret_data = encoder.reverse_transform(data)

        elif self._cont_normalizer == 'stds' or self._cont_normalizer == 'mms11' or self._cont_normalizer == 'stds-pca'\
                or self._cont_normalizer == 'yeo':
            ret_data = encoder.inverse_transform(column_data)

        elif self._cont_normalizer == 'None':
            ret_data = column_data

        # Apply value clipping here: Given an interval, values outside the interval are clipped to the interval edges.
        if self._clip:
            ret_data = np.clip(ret_data, column_transform_info.column_min, column_transform_info.column_max)

        return ret_data

    def _inverse_transform_discrete(self, column_transform_info, column_data):
        ohe = column_transform_info.transform
        data = pd.DataFrame(column_data, columns=list(ohe.get_output_sdtypes()))
        return ohe.reverse_transform(data)[column_transform_info.column_name]

    def inverse_transform(self, data, sigmas=None):
        """Take matrix data and output raw data.

        Output uses the same type as input to the transform function.
        Either np array or pd dataframe.
        """
        st = 0
        recovered_column_data_list = []
        column_names = []
        for column_transform_info in self._column_transform_info_list:
            dim = column_transform_info.output_dimensions
            column_data = data[:, st:st + dim]
            if column_transform_info.column_type == 'continuous':
                recovered_col_data = self._inverse_transform_continuous(column_transform_info, column_data, sigmas, st)
            else:
                recovered_col_data = self._inverse_transform_discrete(column_transform_info, column_data)

            recovered_column_data_list.append(recovered_col_data)
            column_names.append(column_transform_info.column_name)
            st += dim

        recovered_data = np.column_stack(recovered_column_data_list)
        recovered_data = pd.DataFrame(recovered_data, columns=column_names).astype(self._column_raw_dtypes)
        if not self.dataframe:
            recovered_data = recovered_data.to_numpy()

        return recovered_data

    def convert_column_name_value_to_id(self, column_name, value):
        """Get the ids of the given `column_name`."""
        discrete_counter = 0
        column_id = 0
        for column_transform_info in self._column_transform_info_list:
            if column_transform_info.column_name == column_name:
                break
            if column_transform_info.column_type == 'discrete':
                discrete_counter += 1

            column_id += 1

        else:
            raise ValueError(f"The column_name `{column_name}` doesn't exist in the data.")

        ohe = column_transform_info.transform
        data = pd.DataFrame([value], columns=[column_transform_info.column_name])
        one_hot = ohe.transform(data).to_numpy()[0]
        if sum(one_hot) == 0:
            raise ValueError(f"The value `{value}` doesn't exist in the column `{column_name}`.")

        return {
            'discrete_column_id': discrete_counter,
            'column_id': column_id,
            'value_id': np.argmax(one_hot)
        }

    def get_column_transform_info_list(self):
        return self._column_transform_info_list



import os
import sys

import numpy as np
import time

from ctd_gan import ctdGAN

from TabularDataset import TabularDataset
from Tools import set_random_states

num_threads = 1
os.environ['OMP_NUM_THREADS'] = str(num_threads)
np.set_printoptions(linewidth=400, threshold=sys.maxsize)
seed = 1

# dataset_path = '/media/leo/7CE54B377BB9B18B/datasets/Imbalanced/multiclass_continuous/'
dataset_path = 'D:/datasets/Imbalanced/multiclass_continuous/'
datasets = {
    'anemia': {'path': dataset_path + 'anemia.csv', 'categorical_cols': (), 'class_col': 14},
}


# Press the green button in the gutter to run the script.
if __name__ == '__main__':
    set_random_states(seed)
    dataset = datasets['anemia']

    dset = TabularDataset(name='test', class_column=dataset['class_col'],
                          categorical_columns=dataset['categorical_cols'],  random_state=seed)
    dset.load_from_csv(path=dataset['path'])
    dset.display_params()

    x = dset.x_
    y = dset.y_

    t_s = time.time()

    gan = ctdGAN(discriminator=(256, 256), generator=(256, 256), epochs=10, batch_size=100,
                 pac=10, embedding_dim=128, max_clusters=20, cluster_method='kmeans', scaler='mms11',
                 sampling_strategy='create-new', random_state=seed)

    balanced_data = gan.fit_resample(x, y, categorical_columns=dataset['categorical_cols'])

    print("Balanced Data shape:", balanced_data[0].shape)
    # print(balanced_data[0])
    print("Finished in", time.time() - t_s, "sec")



import pathlib
import gzip
import json

import pandas as pd

import seaborn as sns
import matplotlib.pyplot as plt

from sklearn.preprocessing import LabelEncoder
from sklearn.datasets import make_classification
from sklearn.model_selection import cross_validate
from sklearn.metrics import make_scorer
from sklearn.metrics import f1_score, accuracy_score, balanced_accuracy_score, precision_score, recall_score

import warnings

warnings.filterwarnings("ignore")

pd.set_option('display.max_rows', 1000)
pd.set_option('display.max_columns', None)
pd.set_option('display.width', None)
pd.set_option('display.max_colwidth', 400)


class TabularDataset:
    def __init__(self, name, categorical_columns=None, class_column=None, random_state=0):
        """
        Tabular Dataset initializer

        Args:
            name (string): the name of the tabular dataset
            categorical_columns (List or Tuple): A list or tuple with the indices of the categorical columns (if any).
            class_column (int): The index of the column that stores the class of the sample (if any).
            random_state: Controls random number generation. Set this to a fixed integer to get reproducible results.
        """
        self._name = name
        self._random_state = random_state

        self.class_column = class_column
        self.num_classes = 0

        # Raw dataset dimensions - (rows, columns)
        self.num_rows = 0
        self.num_columns = 0

        # Dimensionality is NOT equal to the number of columns.
        # e.g. a single categorical column may have multiple dimensions if one-hot-encoded
        self.dimensionality = 0

        # The raw Dataframe
        self._raw_df = None

        # The Dataframe after some processing (will be used for training)
        self.df_ = None

        # Create one LabelEncoder object for each categorical column
        if categorical_columns is None:
            self.categorical_columns = None
        else:
            self.categorical_columns = list(categorical_columns)
            self._label_encoders = [LabelEncoder() for _ in range(len(self.categorical_columns))]

        self._class_encoder = LabelEncoder()
        self.x_ = None
        self.y_ = None

    # Create synthetic imbalanced datasets with numeric features
    def create_synthetic(self, num_samples=1000, num_classes=2, imb_ratio=(0.5, 0.5)):
        """
        Create a synthetic imbalanced dataset with numeric features.

        Args:
            num_samples: how many data instances to create.
            num_classes: the number of classes of the synthetic data instances.
            imb_ratio: the imbalance ratio of the data (it must be a tuple comprising n_classes elements with sum=1).
        """
        synthetic_dataset = []
        if num_classes == 2:
            synthetic_dataset = make_classification(
                n_samples=num_samples, n_features=2, n_clusters_per_class=1, flip_y=0, n_classes=2, weights=imb_ratio,
                class_sep=0.5, n_informative=2, n_redundant=0, n_repeated=0, random_state=self._random_state)

        elif num_classes == 4:
            synthetic_dataset = make_classification(
                n_samples=num_samples, n_features=2, n_clusters_per_class=1, flip_y=0, n_classes=4, weights=imb_ratio,
                class_sep=1.0, n_informative=2, n_redundant=0, n_repeated=0, random_state=self._random_state)

        self.class_column = 1
        self.num_rows = num_samples
        self.num_classes = num_classes

        x = synthetic_dataset[0]
        y = synthetic_dataset[1]

        self.num_rows = num_samples
        self.num_columns = self.dimensionality = x.shape[1]

        print("Num Samples:", self.num_rows, "\nClass Distribution:")
        for k in range(self.num_classes):
            print("\tClass", k, ":", len(y[y == k]), "samples")

    # Load a dataset from an external CSV file
    def load_from_csv(self, path=''):
        """
        Load a Dataset from a CSV file and compute several basic statistics (number of samples, number of classes,
        input dimensionality). The target variables are Label Encoded.

        Args:
            path: The location of the input CSV file.
        """
        file_extension = pathlib.Path(path).suffix
        if file_extension == '.csv':
            self._raw_df = pd.read_csv(path, skipinitialspace=True, encoding='utf-8',
                                       keep_default_na=True, na_values="<null>")
        else:
            self._raw_df = self.get_df(path)

        # Preprocessing steps:
        # Step 1: Remove rows with missing values.
        self.df_ = self._raw_df.dropna(inplace=False)
        self.df_.reset_index(drop=True, inplace=True)

        # Step 2: Shuffle the dataframe
        self.df_.sample(frac=1)

        self.num_rows = self.df_.shape[0]
        self.num_columns = self.df_.shape[1]

        # Step 3: If a class column exists, put it in the end
        if self.class_column is not None:
            if self.class_column < self.num_columns - 1:
                df_class = self.df_.iloc[:, self.class_column]
                self.df_.drop(self.df_.columns[self.class_column], axis=1, inplace=True)

                self.df_ = pd.concat([self.df_, df_class], axis=1)

                # Adjust the indices of the categorical columns
                if self.categorical_columns is not None:
                    for col in range(len(self.categorical_columns)):
                        if self.categorical_columns[col] > self.class_column:
                            self.categorical_columns[col] -= 1

                self.class_column = self.num_columns - 1

        # Step 4: Change the name of the Dataframe columns
        self.df_.columns = [str(c) for c in range(self.num_columns)]

        # Step 5: Fit_transform the label encoders
        if self.categorical_columns is not None:
            for c in range(len(self.categorical_columns)):
                col = str(self.categorical_columns[c])
                self.df_[col] = self._label_encoders[c].fit_transform(self.df_[col])

        # Step 6: Label Encode the class labels (after stripping the whitespace)
        if pd.api.types.is_string_dtype(self.df_[str(self.class_column)].dtype):
            self.df_[str(self.class_column)] = self.df_[str(self.class_column)].map(str.strip)
        self.df_[str(self.class_column)] = self._class_encoder.fit_transform(self.df_[str(self.class_column)])

        self.x_ = self.df_.iloc[:, 0:self.class_column].to_numpy()
        self.y_ = self.df_.iloc[:, self.class_column].to_numpy()

        # Useful quick reference statistics
        self.num_classes = len(self.df_.iloc[:, self.class_column].unique())
        self.dimensionality = self.x_.shape[1]

        # print("Num Samples:", self.num_rows, "\nClass Distribution:")
        # for k in range(self.num_classes):
        #    print("\tClass", k, ":", len(self.y_[self.y_ == k]), "samples")

    # Get dummies - onehot encode the categorical columns
    def get_dummies(self):
        if self.categorical_columns is not None:
            str_cat_columns = [str(c) for c in self.categorical_columns]
            transformed_df = pd.get_dummies(data=self.df_, columns=str_cat_columns, dtype=int)

            return transformed_df

    # Display the basic dataset parameters
    def display_params(self):
        """
        Display the basic dataset parameters.
        """
        print("===================================================================================================")
        print("Dataset:", self._name)
        print("Rows:", self.num_rows)
        print("Columns:", self.num_columns, ", Categorical:", len(self.categorical_columns))
        print("Encoded data Dimensions:", self.dimensionality)
        print("Classes:", self.num_classes, "- Class Distribution:")
        for k in range(self.num_classes):
            print("\t\tClass", k, ":", len(self.y_[self.y_ == k]), "samples")
        print("===================================================================================================")

    def plot(self, dim1=0, dim2=1):
        """
        Uses a 2-dimensional space to plot two features of the dataset and color the data points according to their
        class.
        """
        sns.scatterplot(x=self.x_[:, dim1], y=self.x_[:, dim2], hue=self.y_, legend=False)
        plt.show()

    # Bring balance to the dataset by applying a pipeline with an over-/under-sampling technique. Then, train and test
    # a classifier with cross-validation. A pipeline object is required.
    def cross_val(self, estimator, num_folds, num_threads, classifier_str, sampler_str, order):
        """
        Applies a classification pipeline to the dataset, and then, it evaluates the classification performance by using
        cross validation.

        Args:
            estimator: A multi-stage Pipeline object. The Pipeline must include a classifier at its final stage.
            num_folds: Number of folds for cross validation.
            num_threads: The number of parallel threads for cross validation.
            classifier_str: A custom string that describes the Classifier in the Pipeline.
            sampler_str: A custom string that describes the other stages of the Pipeline.
            order: An assistant variable for enumerating the returned results.

        Returns:
            Two objects that store all cross validation results.
        """
        scorers = {
            'accuracy': make_scorer(accuracy_score),
            'balanced_accuracy': make_scorer(balanced_accuracy_score),
            'f1': make_scorer(f1_score, average='weighted'),
            'precision': make_scorer(precision_score, average='weighted'),
            'recall': make_scorer(recall_score, average='weighted'),
        }

        # cross_validate uses Stratified kFold when cv is int
        cv_results = cross_validate(estimator, self.x_, self.y_, cv=num_folds, scoring=scorers,
                                    return_train_score=False, return_estimator=False, n_jobs=num_threads,
                                    error_score='raise')

        results_list = []
        for key in cv_results.keys():
            for f in range(num_folds):
                lst = [self._name, f + 1, sampler_str, classifier_str, key, cv_results[key][f]]
                results_list.append(lst)

        cv_results['dataset'] = self._name
        cv_results['sampler'] = sampler_str
        cv_results['classifier'] = classifier_str

        mean_results_list = [
            classifier_str,
            sampler_str,
            self._name,
            cv_results['test_accuracy'].mean(),
            cv_results['test_accuracy'].std(),
            cv_results['test_balanced_accuracy'].mean(),
            cv_results['test_balanced_accuracy'].std(),
            cv_results['test_sensitivity'].mean(),
            cv_results['test_sensitivity'].std(),
            cv_results['test_specificity'].mean(),
            cv_results['test_specificity'].std(),
            cv_results['test_f1'].mean(),
            cv_results['test_f1'].std(),
            cv_results['test_precision'].mean(),
            cv_results['test_precision'].std(),
            cv_results['test_recall'].mean(),
            cv_results['test_recall'].std(),
            cv_results['fit_time'].mean(),
            order
        ]

        return results_list, mean_results_list

    # Return the dataframe
    def get_data(self):
        """
        Return the Dataframe of the dataset
        :return: A Pandas Dataframe with the stored data
        """
        return self.df_

    def get_df(self, path):
        i = 0
        df = {}
        for d in self.parse(path):
            df[i] = d
            i += 1
        return pd.DataFrame.from_dict(df, orient='index')

    @staticmethod
    def parse(path):
        g = gzip.open(path, 'rb')
        for lt in g:
            yield json.loads(lt)



import numpy as np
import torch

import gc
import contextlib


def set_random_states(manual_seed):
    """Initialize the random number generators of NumPy, PyTorch, and PyTorch CUDA by passing the input seed.

    Args:
        manual_seed: An integer to be passed to the random number generators.
    """
    np.random.seed(manual_seed)

    if manual_seed is None:
        torch.manual_seed(0)
        torch.cuda.manual_seed(0)
    else:
        torch.manual_seed(manual_seed)
        torch.cuda.manual_seed(manual_seed)

    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False


def get_random_states():
    """Retrieves the current states of randomness of NumPy, PyTorch, and PyTorch CUDA.

    Returns:
        Three states of randomness for NumPy, PyTorch, and PyTorch CUDA respectively.
    """
    np_random_state = np.random.get_state()
    torch_random_state = torch.random.get_rng_state()
    if torch.cuda.is_available():
        cuda_random_state = torch.cuda.random.get_rng_state()
    else:
        cuda_random_state = None

    return np_random_state, torch_random_state, cuda_random_state


def reset_random_states(np_random_state, torch_random_state, cuda_random_state):
    """Sets the current states of randomness of NumPy, PyTorch, and PyTorch CUDA.

    Args:
        np_random_state: The state at which the NumPy random generator will be set.
        torch_random_state: The state at which the PyTorch random generator will be set.
        cuda_random_state: The state at which the PyTorch CUDA random generator will be set.
    """
    np.random.set_state(np_random_state)
    torch.random.set_rng_state(torch_random_state)

    if torch.cuda.is_available():
        torch.cuda.random.set_rng_state(cuda_random_state)
        torch.cuda.empty_cache()

    gc.collect()


@contextlib.contextmanager
def ct_set_random_states(seed, set_model_random_state):
    """Context manager for managing the random state.

    Args:
        seed (int or tuple):
            The random seed or a tuple of (numpy.random.RandomState, torch.Generator).
        set_model_random_state (function):
            Function to set the random state on the model.
    """
    original_np_state = np.random.get_state()
    original_torch_state = torch.get_rng_state()

    random_np_state, random_torch_state = seed

    np.random.set_state(random_np_state.get_state())
    torch.set_rng_state(random_torch_state.get_state())

    try:
        yield
    finally:
        current_np_state = np.random.RandomState()
        current_np_state.set_state(np.random.get_state())
        current_torch_state = torch.Generator()
        current_torch_state.set_state(torch.get_rng_state())
        set_model_random_state((current_np_state, current_torch_state))

        np.random.set_state(original_np_state)
        torch.set_rng_state(original_torch_state)


def random_state(function):
    """Set the random state before calling the function.

    Args:
        function (Callable): The function to wrap around.
    """

    def wrapper(self, *args, **kwargs):
        if self.random_states is None:
            return function(self, *args, **kwargs)

        else:
            with ct_set_random_states(self.random_states, self.set_random_state):
                return function(self, *args, **kwargs)

    return wrapper



import torch
import torch.nn as nn


# import DeepCoreML.paths as paths
class Critic(nn.Module):
    """Discriminator for ctGAN."""

    def __init__(self, input_dim, discriminator_dim, pac=10):
        super().__init__()
        dim = input_dim * pac
        self._pac = pac
        self._pac_dim = dim
        seq = []
        for item in list(discriminator_dim):
            seq += [nn.Linear(dim, item), nn.LeakyReLU(0.2), nn.Dropout(0.5)]
            dim = item

        seq += [nn.Linear(dim, 1)]
        self._seq = nn.Sequential(*seq)

    def calc_gradient_penalty(self, real_data, fake_data, device='cpu', lambda_=10):
        """Compute the gradient penalty. From the paper on improved WGAN training."""
        alpha = torch.rand(real_data.size(0) // self._pac, 1, 1, device=device)
        alpha = alpha.repeat(1, self._pac, real_data.size(1))
        alpha = alpha.view(-1, real_data.size(1))

        interpolates = alpha * real_data + ((1 - alpha) * fake_data)

        disc_interpolates = self(interpolates)

        gradients = torch.autograd.grad(
            outputs=disc_interpolates, inputs=interpolates,
            grad_outputs=torch.ones(disc_interpolates.size(), device=device),
            create_graph=True, retain_graph=True, only_inputs=True
        )[0]

        gradients_view = gradients.view(-1, self._pac * real_data.size(1)).norm(2, dim=1) - 1
        gradient_penalty = (gradients_view ** 2).mean() * lambda_

        return gradient_penalty

    def forward(self, x):
        """Apply the Discriminator to the `input`."""
        assert x.size()[0] % self._pac == 0
        return self._seq(x.view(-1, self._pac_dim))


class Residual(nn.Module):
    """Residual layer for the CTGAN."""

    def __init__(self, i, o):
        super(Residual, self).__init__()
        self.fc = nn.Linear(i, o)
        self.bn = nn.BatchNorm1d(o)
        self.relu = nn.ReLU()

    def forward(self, input_):
        """Apply the Residual layer to the `input_`."""
        out = self.fc(input_)
        out = self.bn(out)
        out = self.relu(out)
        return torch.cat([out, input_], dim=1)


class Generator(nn.Module):
    """Generator for ctGAN and ctdGAN"""

    def __init__(self, embedding_dim, architecture, data_dim):
        super().__init__()
        dim = embedding_dim
        seq = []
        for item in list(architecture):
            seq += [Residual(dim, item)]
            dim += item
        seq.append(nn.Linear(dim, data_dim))
        self.seq = nn.Sequential(*seq)

    def forward(self, input_):
        """Apply the Generator to the `input_`."""
        data = self.seq(input_)
        return data



import numpy as np

from sklearn.ensemble import IsolationForest

from sklearn.cluster import KMeans, AgglomerativeClustering
from sklearn.mixture import GaussianMixture
from sklearn.preprocessing import StandardScaler, MinMaxScaler, OneHotEncoder, PowerTransformer
from sklearn.compose import ColumnTransformer

from joblib import Parallel, delayed


class ctdCluster:
    """A typical cluster for ctdGAN."""
    def __init__(self, label=None, scaler=None, embedding_dim=32, clip=False,
                 continuous_columns=(), discrete_columns=(), random_state=0):
        """
        ctdCluster initializer. A typical cluster for ctdGAN.

        Args:
            label: The cluster's label
            scaler (string): A descriptor that defines a transformation on the cluster's data. Values:

              * '`None`'  : No transformation takes place; the data is considered immutable
              * '`stds`'  : Standard scaler
              * '`mms01`' : Min-Max scaler in the range (0,1)
              * '`mms11`' : Min-Max scaler in the range (-1,1) - so that data is suitable for tanh activations
              * 'yeo': Yeo-Johnson power transformer.
            embedding_dim (int): The dimensionality of the latent space (for the probability distribution)
            continuous_columns (tuple): The continuous columns in the input data
            discrete_columns (tuple): The columns in the input data that contain categorical variables
            clip (bool): If 'True' the reconstructed data will be clipped to their original minimum and maximum values.
            random_state (int): Seed the random number generators. Use the same value for reproducible results.
        """
        self._label = label
        self._embedding_dim = embedding_dim
        self._continuous_columns = continuous_columns
        self._discrete_columns = discrete_columns
        self._clip = clip
        self._random_state = random_state

        self._num_samples = 0
        self._min = None
        self._max = None

        self.class_distribution_ = None

        # Define the Data Transformation Model for the continuous columns
        if len(continuous_columns) > 0:
            if scaler == 'stds':
                self._scaler = StandardScaler()
            elif scaler == 'mms01':
                self._scaler = MinMaxScaler(feature_range=(0, 1))
            elif scaler == 'mms11':
                self._scaler = MinMaxScaler(feature_range=(-1, 1))
            elif scaler == 'yeo':
                self._scaler = PowerTransformer(method='yeo-johnson', standardize=True)
            else:
                self._scaler = None
        else:
            self._scaler = None

    def fit(self, x, y=None, num_classes=0):
        """
            Compute cluster statistics and fit the selected Data Transformer.

        Args:
            x (NumPy array): The data to be transformed
            y (NumPy array): If the data has classes, pass them here. The ctdGAN will be trained for resampling.
            num_classes: The distinct number of classes in `y`.
        """
        self._num_samples = x.shape[0]

        # Min/Max column values are used for clipping.
        self._min = [np.min(x[:, i]) for i in self._continuous_columns]
        self._max = [np.max(x[:, i]) for i in self._continuous_columns]

        self.class_distribution_ = np.zeros(num_classes)
        if y is not None:
            unique_classes = np.unique(y, return_counts=True)
            n = 0
            for uv in unique_classes[0]:
                self.class_distribution_[uv] = unique_classes[1][n]
                n += 1

        if self._scaler is not None:
            x_cont = x[:, self._continuous_columns]
            self._scaler.fit(x_cont)

    def transform(self, x):
        if self._scaler is not None:
            # print("Before Transformation:\n", x)
            x_cont = x[:, self._continuous_columns]
            transformed = self._scaler.transform(x_cont)
            # print("After Transformation of Continuous Cols:\n", transformed)

            for d_col in self._discrete_columns:
                transformed = np.insert(transformed, d_col, x[:, d_col], axis=1)
            # print("After Transformation & concat with Discrete Cols:\n", transformed)
            # exit()
            return transformed
        else:
            return x

    def fit_transform(self, x):
        """Transform the sample vectors by applying the transformation function of `self._scaler`. In fact, this
        is a simple wrapper for the `fit_transform` function of `self._scaler`.

        `self._scaler` may implement a `Pipeline`.

        Returns:
            The transformed data.
        """
        self.fit(x)
        return self._scaler.transform(x)

    def inverse_transform(self, x):
        """
        Inverse the transformation that has been applied by `self.fit_transform()`. In fact, this is a wrapper for the
        `inverse_transform` function of `self._scaler`, followed by a filter that clips the returned values.

        Args:
            x: The input data to be reconstructed (NumPy array).
            x: The input data to be reconstructed (NumPy array).

        Returns:
            The reconstructed data.
        """

        if self._scaler is not None:
            x_cont = x[:, self._continuous_columns]
            reconstructed_data = self._scaler.inverse_transform(x_cont)

            if self._clip:
                np.clip(reconstructed_data, self._min, self._max, out=reconstructed_data)

            for d_col in self._discrete_columns:
                reconstructed_data = np.insert(reconstructed_data, d_col, x[:, d_col], axis=1)
        else:
            # reconstructed_data = x.copy()
            reconstructed_data = x[:, 0:(x.shape[1] - 2)]

        return reconstructed_data

    def display(self):
        """
        Display useful cluster properties.
        """
        print("\t--- Cluster ", self._label, "-----------------------------------------")
        print("\t\t* Num Samples: ", self._num_samples)
        print("\t\t* Class Distribution: ", self.class_distribution_)
        print("\t\t* Min values per column:", self._min)
        print("\t\t* Max values per column:", self._max)
        print("\t---------------------------------------------------------\n")

    def get_label(self):
        return self._label

    def get_num_samples(self, c=None):
        if c is None:
            return self._num_samples
        else:
            if len(self.class_distribution_) > 0:
                return self.class_distribution_[c]
            else:
                return -1

    def set_label(self, v):
        self._label = v


class ctdClusterer:
    """  ctdGAN data preprocessor.

    It partitions the real space into clusters; then, it transforms the data of each cluster.
    """
    def __init__(self, cluster_method='kmeans', max_clusters=10, scaler=None, samples_per_class=(), embedding_dim=32,
                 continuous_columns=(), discrete_columns=(), random_state=0):
        """
        Initializer

        Args
            cluster_method (str): The clustering algorithm to apply. Supported values:
              * kmeans: K-Means
              * hac: Hierarchical Agglomerative Clustering
              * gmm: Gaussian Mixture Model

            max_clusters (int): The maximum number of clusters to create
            scaler (string): A descriptor that defines a transformation on the cluster's data. Supported values:

              * '`None`'  : No transformation takes place; the data is considered immutable
              * '`stds`'  : Standard scaler
              * '`mms01`' : Min-Max scaler in the range (0,1)
              * '`mms11`' : Min-Max scaler in the range (-1,1) - so that data is suitable for tanh activations
            embedding_dim (int): The dimensionality of the latent space (for the probability distribution)
            continuous_columns (tuple): The continuous columns in the input data
            discrete_columns (tuple): The columns in the input data that contain categorical variables
            samples_per_class (List or tuple of integers): Contains the number of samples per class
            random_state: Seed the random number generators. Use the same value for reproducible results.
        """
        self._cluster_method = cluster_method
        if cluster_method not in ['kmeans', 'hac', 'gmm']:
            self._cluster_method = 'kmeans'

        self._scaler = scaler
        if scaler not in ['None', 'stds', 'mms01', 'mms11', 'yeo']:
            self._scaler = 'stds'

        self._max_clusters = max_clusters
        self._random_state = random_state
        self._embedding_dim = embedding_dim
        self._continuous_columns = continuous_columns
        self._discrete_columns = discrete_columns

        self.num_clusters_ = 0
        self.clusters_ = []
        self.cluster_labels_ = None
        self.probability_matrix_ = None
        self.imbalance_matrix_ = None
        self._samples_per_class = samples_per_class

    def perform_clustering(self, x_train, y_train, num_classes, pac):
        """

        Args:
            x_train: Training data
            y_train: The classes of the training samples
            num_classes: The number of distinct classes of the training data
            pac (int): The number of samples to group together as input to the Critic.

        Returns:
            Transformed data
        """
        # Prepare the data for clustering:
        # 1) MinMax scale the continuous variables, and 2) OneHotEncode the discrete variables.
        column_transformer = ColumnTransformer([
            ("mms", MinMaxScaler(), self._continuous_columns),
            ("ohe", OneHotEncoder(), self._discrete_columns)
        ], sparse_threshold=0)
        x_scaled = column_transformer.fit_transform(x_train)

        # Find the optimal number of clusters (best_k).
        # Perform multiple executions and pick the one that produces the minimum scaled inertia.
        # jobs = 0.2 * multiprocessing.cpu_count()
        jobs = -1
        k_range = range(2, self._max_clusters)
        scores = Parallel(n_jobs=jobs)(delayed(self._run_clustering)(x_scaled, k) for k in k_range)
        best = min(scores, key=lambda score_tuple: score_tuple[1])
        self.num_clusters_ = best[0]
        best_cov_type = best[2]

        # After the optimal number of clusters best_k has been determined, execute one last k-Means with best_k clusters
        if self._cluster_method == 'hac':
            cluster_method = AgglomerativeClustering(n_clusters=self.num_clusters_)
        elif self._cluster_method == 'kmeans':
            cluster_method = KMeans(n_clusters=self.num_clusters_, n_init='auto', random_state=self._random_state)
        elif self._cluster_method == 'gmm':
            cluster_method = GaussianMixture(n_components=self.num_clusters_, covariance_type=best_cov_type,
                                             random_state=self._random_state)
        else:
            cluster_method = KMeans(n_clusters=self.num_clusters_, n_init='auto', random_state=self._random_state)

        self.cluster_labels_ = cluster_method.fit_predict(x_scaled)

        # Partition the dataset and create the appropriate Cluster objects.
        transformed_data = None
        for u in range(self.num_clusters_):
            x_u = x_train[self.cluster_labels_ == u, :]
            y_u = y_train[self.cluster_labels_ == u]

            cluster = ctdCluster(label=u, scaler=self._scaler,
                                 clip=True, embedding_dim=self._embedding_dim,
                                 continuous_columns=self._continuous_columns, discrete_columns=self._discrete_columns,
                                 random_state=self._random_state)
            cluster.fit(x_u, y_u, len(self._samples_per_class))

            # print(x_u)
            x_transformed = cluster.transform(x_u)
            # print(x_transformed)
            # exit()
            cluster_labels = (u * np.ones(y_u.shape[0])).reshape(-1, 1)
            class_labels = np.array(y_u).reshape(-1, 1)

            if u == 0:
                transformed_data = np.concatenate((x_transformed, cluster_labels, class_labels), axis=1)
            else:
                concat = np.concatenate((x_transformed, cluster_labels, class_labels), axis=1)
                transformed_data = np.concatenate((transformed_data, concat))

            self.clusters_.append(cluster)

        # Forge the probability matrix; Each element (i,j) stores the conditional probability
        # P(cluster==u | class=y) = P( (class==y) AND (cluster==u) ) / P(class==y)
        if num_classes > 1:
            self.probability_matrix_ = np.zeros((num_classes, self.num_clusters_))
            self.imbalance_matrix_ = np.zeros((num_classes, self.num_clusters_))
            for c in range(num_classes):
                for u in range(self.num_clusters_):
                    cluster = self.clusters_[u]
                    # print("\t Cluster:", comp, "- Samples:", cluster.get_num_samples(),
                    #      "(", cluster.get_num_samples(c), "from class", c, ")")
                    self.probability_matrix_[c][u] = cluster.get_num_samples(c) / self._samples_per_class[c]
                    self.imbalance_matrix_[c][u] = cluster.get_num_samples(c)

        # Pad the dataset to align with the pac parameter (Create integral number of groups of pac samples).
        dataset_rows = transformed_data.shape[0]
        if dataset_rows % pac != 0:
            required_samples = pac * (dataset_rows // pac + 1) - dataset_rows
            random_samples = transformed_data[np.random.randint(0, dataset_rows, (required_samples,))]
            transformed_data = np.vstack((transformed_data, random_samples))

        # Shuffle the dataset
        np.random.shuffle(transformed_data)

        return transformed_data

    def remove_majority_outliers(self, x_train, y_train):
        num_samples = x_train.shape[1]
        majority_class = np.argmax(self._samples_per_class)
        maj_samples = np.array([x_train[s, :] for s in range(num_samples) if y_train[s] == majority_class])

        # Use an Isolation Forest to detect the outliers. The predictions array marks them with -1
        outlier_detector = IsolationForest(random_state=self._random_state)
        outlier_detector.fit(maj_samples)
        predictions = outlier_detector.predict(maj_samples)

        # Copy all the minority samples to the cleaned dataset
        x_clean = np.array([x_train[s, :] for s in range(num_samples) if y_train[s] != majority_class])
        y_clean = np.array([y_train[s] for s in range(num_samples) if y_train[s] != majority_class])

        # Copy the majority samples that are not outliers to the cleaned dataset
        for s in range(maj_samples.shape[0]):
            if predictions[s] == 1:
                x_clean = np.append(x_clean, [maj_samples[s, :]], axis=0)
                y_clean = np.append(y_clean, [majority_class], axis=0)

        # (x_clean, y_clean) is the new dataset without the outliers
        # print("Clean Dataset Shape:", x_clean.shape)

    def _run_clustering(self, scaled_data, num_clusters, alpha_k=0.02):
        """
        Args:
        scaled_data: matrix
            scaled data. rows are samples and columns are features for clustering
        max_clusters: int
            current k for applying KMeans
        alpha_k: float
            manually tuned factor that gives penalty to the number of clusters

        Returns:
            scaled_inertia: float
                scaled inertia value for current k
        """

        ret_val = 0
        cov_type = 'None'
        inertia_o = np.square((scaled_data - scaled_data.mean(axis=0))).sum()

        if self._cluster_method == 'hac':
            hac = AgglomerativeClustering(n_clusters=num_clusters, compute_distances=True)
            hac.fit(scaled_data)

            inertia = 0
            for u in range(num_clusters):
                cluster_data = scaled_data[hac.labels_ == u, :]
                inertia += np.square((cluster_data - cluster_data.mean(axis=0))).sum()
            ret_val = inertia / inertia_o + alpha_k * num_clusters

        elif self._cluster_method == 'kmeans':
            kmeans = KMeans(n_clusters=num_clusters, random_state=self._random_state, n_init='auto')
            kmeans.fit(scaled_data)
            ret_val = kmeans.inertia_ / inertia_o + alpha_k * num_clusters

        elif self._cluster_method == 'gmm':
            min_bic = 10 ** 9
            for cov in ['spherical', 'tied', 'diag', 'full']:
                gmm = GaussianMixture(n_components=num_clusters, covariance_type=cov, random_state=self._random_state)
                gmm.fit(scaled_data)
                bic_score = gmm.bic(scaled_data)
                if bic_score < min_bic:
                    min_bic = bic_score
                    cov_type = cov
            ret_val = min_bic

        return num_clusters, ret_val, cov_type

    def display(self):
        print("Num Clusters: ", self.num_clusters_)
        for i in range(self.num_clusters_):
            self.clusters_[i].display()

    def get_cluster(self, i):
        return self.clusters_[i]

    def get_cluster_center(self, i):
        return self.clusters_[i].center_
